%%%%%%% In the name of Allah The Rahman The Raheem %%%%%%%

% Last change: Mar 2, 2016
% Code to estimate channels in a massive MIMO system using the method
% presented in paper:
% M. Masood, L. H. Afify and T. Y. Al-Naffouri, "Efficient Coordinated 
% Recovery of Sparse Channels in Massive MIMO," in IEEE Transactions on 
% Signal Processing, vol. 63, no. 1, pp. 104-118, Jan.1, 2015.
% 
% Copyright (c) Mudassir Masood, Tareq Y. Al-Naffouri, 2016
% 
% Coded by: Mudassir Masood and Tareq Y. Al-Naffouri, KAUST, Saudi Arabia.
% E-mail: mudassir.masood@kaust.edu.sa, tareq.alnaffouri@kaust.edu.sa
%

function [Results]=demo_massiveMIMO

%Initialize required variables. Read comments therein for details
[mMc, Mflags, SimPar, Rel]=InitializationsMIMO;





% Generating channels for simulations
[~, zf]=GenerateMIMOChannels(Mflags.SV, mMc);
% randomly generating pilot locations
IntVar.pilots_locations= randperm(mMc.N_subcarriers, mMc.K);
% Generating simulation data
[OFDMS, ForComparison]= GenerateSimulationData(mMc, SimPar.SNRdB, zf, IntVar.pilots_locations);
% Preparing the necessary variables to be used for detecting most-significant tap locations at each antenna location
[IntVar] = PrepareIntVar_struct(IntVar, mMc, OFDMS);
% Step 1
[DFSC]=FindDataForScoreCalculation_struct(IntVar, mMc, Mflags, OFDMS);
% Step 2
Rel.TS=mMc.N_subcarriers;
[~, RecChan, ~] = mMIMO_mainfunc(mMc, DFSC, OFDMS, Mflags, SimPar, IntVar, Rel);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute NMSE in channel estimate at each antenna
% each element of NMSE matrix contains NMSE in the estimation of the
% channels at that particular antenna location
NMSE = inf*ones(mMc.Nx, mMc.My);
for index_X=1:mMc.Nx
    for index_Y=1:mMc.My
        NMSE(index_X,index_Y) = (norm(RecChan.h_recovered_rel(:,index_Y,index_X) - ForComparison.h(:,index_Y,index_X))^2)/(norm(ForComparison.h(:,index_Y,index_X))^2);
    end
end

%Bit Error calculation
berrors1=0;
berrors2=0;
inds_minus_pilots=setdiff(1:mMc.N_subcarriers,IntVar.pilots_locations);
for index_X=1:mMc.Nx
    for index_Y=1:mMc.My
        
        %bit error in estimated channel data recovery
        data_eq3 = OFDMS.Y(:,index_Y,index_X)./RecChan.h_recovered_rel_f(:,index_Y,index_X);
        data_eq3(IntVar.pilots_locations)=ForComparison.X(IntVar.pilots_locations);
        x_f_unenhanced=qammod(qamdemod(data_eq3,mMc.Q),mMc.Q);
        hIntToBit = comm.IntegerToBit(log2(mMc.Q));
        bitData2 = step(hIntToBit,qamdemod(x_f_unenhanced(inds_minus_pilots),mMc.Q));
        hIntToBit = comm.IntegerToBit(log2(mMc.Q));
        bitData_true = step(hIntToBit,qamdemod(ForComparison.X(inds_minus_pilots),mMc.Q));

        hErrorCalc = comm.ErrorRate;
        berVec = step(hErrorCalc,bitData2,bitData_true);
        berrors2 = berrors2+ berVec(1);
        
        %bit error in known channel data recovery
        data_eq3 = OFDMS.Y(:,index_Y,index_X)./ForComparison.h_cap(:,index_Y,index_X);
        data_eq3(IntVar.pilots_locations)=ForComparison.X(IntVar.pilots_locations);
        x_f_unenhanced=qammod(qamdemod(data_eq3,mMc.Q),mMc.Q);
        hIntToBit = comm.IntegerToBit(log2(mMc.Q));
        bitData1 = step(hIntToBit,qamdemod(x_f_unenhanced(inds_minus_pilots),mMc.Q));
        
        hErrorCalc = comm.ErrorRate;
        berVec = step(hErrorCalc,bitData1,bitData_true);
        berrors1 = berrors1+ berVec(1);
    end
end



fprintf('\n__________________________________________');
fprintf('\nMean NMSE = %f', mean(NMSE(:)));
fprintf('\nBit errors (known channel) = %f', berrors1);
fprintf('\nBit errors (estimated channel) = %f\n', berrors2);
fprintf('\n__________________________________________');


Results.berrors1=berrors1;
Results.berrors2=berrors2;
Results.NMSE=NMSE;


end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [mMc, Mflags, SimPar, Rel]=InitializationsMIMO


%% initializations

mMc.N_subcarriers = 256; %total number of subcarriers
mMc.L = 64; %no. of channel taps (channel length)
mMc.K = 8; %no. of pilot tones
mMc.Q = 4; %QAM constellation size
mMc.n = 3; %sparsity of L length channel
mMc.My = 5; %antenna dimension in y-direction
mMc.Nx = 5; %antenna dimension in x-direction (i.e., total number of columns)
mMc.D1 = 1; %value of D passed to SABMP when finding marginal posteriors
mMc.D2 = 5; %value of D for the SABMP recovery using pilot tones only
mMc.D3 = 2; %value of D for the SABMP recovery using pilots + reliables
mMc.norm_factor=sqrt(1/mMc.K);
mMc.p2 = mMc.n/mMc.L; %sparsity rate
mMc.p1 = mMc.p2*ones(mMc.L,mMc.My,mMc.Nx);  % initial vector of Bernoulli probabilities
mMc.Delta = 3;
mMc.topV = mMc.n + mMc.Delta; %this many indices are picked as possible support based on the top values of scores
mMc.Four_neighbors = Find4NeighborsIndices(mMc.My,mMc.Nx); %returns the (x,y) coordinates of four neighbors for all elements in the grid. returns a cell containing vectors with (x,y) coordinates of each element in the two dimensional antenna array grid.
mMc.No_of_RelTones_Eb = mMc.N_subcarriers/1; %number of reliable tones to be picked in Ebrahim's technique. Note that these many reliable tones will be picked and then subjected to data intersection to decide the final set of reliable tones to be used in channel estimation.

Mflags.PB = 0; 
Mflags.SV = 0; % SV=1 --> space-varying impulse response. SV=0 --> space-invariant impulse response.

SimPar.SNRdB = 15;

SimPar.max_p_iter = 2;
SimPar.NeighborSearchLevel = 1;
SimPar.NextTierNeighborMethod = 2; % =2 using just 4-neighbors at a time

Rel.C_F_var = FindStatsEbRelCalculation_struct(mMc); % this function is used to find some statistics to be used by the reliability calculation code of Ebrahim.
Rel.R_min = 1/3; % This is parameter for Ebrahim's technique. 
Rel.TS = mMc.No_of_RelTones_Eb;
Rel.rel_rank_thresh = 1;

end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




function [OFDMS, ForComparison]= GenerateSimulationData(mMc, SNRdB, zf, pilots_locations)

N_subcarriers = mMc.N_subcarriers;
L = mMc.L;
K = mMc.K;
Q = mMc.Q;
My = mMc.My;
Nx = mMc.Nx;

% INPUTS:
% N_subcarriers = total number of subcarriers
% L = number of channel taps
% K = number of pilot tones
% Q = constellation size of QAM
% SNRdB = desired SNR in dB
% zf = The channel generated by our antenna array impulse response generation function
% My and Nx determine the array dimensions in y and x directions
% respectively. For a linear array one of thee quantities should be 1.
% pilots_locations = pilot locations
%
% OUTPUTS:
% pilots_locations = pilot tone indices
% y = time-domain received signal
% Y = freq domain received signal
% X_pilot = QPSK modulated data on pilot tones
% F_pilot = partial Fourier matrix corresponding to pilot locations
% F = DFT matrix of dimension N_subcarriers x N_subcarriers
% FH = conjugate transpose of F.
% sig2w = noise variance
% h = channel impulse response (which is basically nothing but a
% rearrangement of zf sent as input)
% X = modulated data (freq domain)

dataseq=randi(Q, N_subcarriers, 1)-1; %data generated between [0 Q-1]
X=qammod(dataseq,Q); % QAM modulation of data

F=(1/sqrt(N_subcarriers))*dftmtx(N_subcarriers); % DFT matrix of dim : N_subcarriers x N_subcarriers
FH=F';
F_N_subcarriers_by_L=F(:,1:L);
F_pilot=F_N_subcarriers_by_L(pilots_locations,:); %random placement of pilots

pilot_4QAM=randi(4,K,1)-1;
X_pilot=pskmod(pilot_4QAM, 4);
X(pilots_locations)=X_pilot;

x=FH*X; %time-domain signal

h=zeros(L,My,Nx); %channel impulse response
h_cap=zeros(N_subcarriers,My,Nx); %channel frequency response
y=zeros(N_subcarriers,My,Nx);
Y=zeros(N_subcarriers,My,Nx);
sig2w=zeros(My,Nx);

for index_X=1:Nx
    for index_Y=1:My
        h(:,index_Y,index_X)=squeeze(zf(index_Y, index_X,:)); %*[1 + (2*rand(1,1)-1)*1i];  %% the L-tap h vector...we need to correct the way the complex impulse response is being generated.
        h1 = h(:,index_Y,index_X);
        h_cap(:,index_Y,index_X) = sqrt(N_subcarriers) * F * [h1; zeros(N_subcarriers-length(h1), 1)]; % Fourier transfor i.e. frequency responce
        D = diag(h_cap(:,index_Y,index_X), 0); % the channel diagonal matrix
        H = FH*D*F; % the circulant channel matrix
        w = [randn(N_subcarriers,1) randn(N_subcarriers,1)]*[1;1i]/sqrt(2);         % AWGN samples
        sig2w1=var(w);
        w=w/(sqrt(sig2w1));
        sig2w(index_Y, index_X) = norm(H*x)^2/N_subcarriers*10^(-SNRdB/10);   % Noise variance from SNR and signal realization
        w=w*sqrt(sig2w(index_Y, index_X));
        y(:,index_Y,index_X) = H*x + w;    %time-domain received signal
        Y(:,index_Y, index_X) = F*y(:,index_Y,index_X);
    end
end

OFDMS.y = y;
OFDMS.Y = Y;
OFDMS.X_pilot = X_pilot;
OFDMS.F_pilot = F_pilot;
OFDMS.F = F;
OFDMS.FH = FH;
OFDMS.F_N_subcarriers_by_L = F_N_subcarriers_by_L;
OFDMS.sig2w = sig2w;

ForComparison.h = h;
ForComparison.X = X;
ForComparison.H = H;
ForComparison.h_cap = h_cap;
end





